<?xml version="1.0"?>

<!--
=================================================================================================
Copyright 2012 Twitter, Inc.
_________________________________________________________________________________________________
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this work except in compliance with the License.
You may obtain a copy of the License in the LICENSE file, or at:

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
=================================================================================================
-->

<!-- generated by pants! -->
<factorypath>
  {{#factorypath.jarpaths}}
  <factorypathentry kind="WKSPJAR" id="/{{factorypath.project_name}}/{{jar}}"
                    enabled="true"
                    runInBatchMode="false"/>
  {{/factorypath.jarpaths}}
</factorypath>
