{{! Chrome for reporting console. }}
<html>
<head>
  <title>Pants</title>
  <link rel="stylesheet" type="text/css" href="/assets/css/pants.css">
  <link rel="stylesheet" type="text/css" href="/assets/css/font-awesome.css">
  <script type="text/javascript" src="/assets/js/jquery-1.8.3.min.js"></script>
  <script type="text/javascript" src="/assets/js/pants.js"></script>
</head>
<body>

<div id="container">
<div id="top">
<h1 class="bigtext">Pants</h1>
</div>

<div id="middle">
<div id="menu">
<ul id="nav">
  <li><a href="/runs/"><span>Pants Runs</span></a></li>
  <li><a href="/browse/"><span>Browse Codebase</span></a></li>
</ul>
</div>

<div id="main-content">
{{#include}}{{content_template}}{{/include}}
</div>
</div>

<div id="bottom">
</div>

</div>

<div id="sticky-footer">
<div id="sticky-footer-content">
    Footer
</div>
</div>
</body>
</html>
