<?xml version="1.0"?>

<!--
=================================================================================================
Copyright 2011 Twitter, Inc.
_________________________________________________________________________________________________
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this work except in compliance with the License.
You may obtain a copy of the License in the LICENSE file, or at:

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
=================================================================================================
-->

<!-- generated by pants! -->
<project version="4">
  <component name="CompilerConfiguration">
    <option name="DEFAULT_COMPILER" value="Javac"/>
    <resourceExtensions/>
    <wildcardResourcePatterns>
      {{#project.resource_extensions}}
      <entry name="?*{{.}}"/>
      {{/project.resource_extensions}}
    </wildcardResourcePatterns>
    <annotationProcessing enabled="false" useClasspath="true"/>
  </component>

  <component name="JavacSettings">
    <option name="ADDITIONAL_OPTIONS_STRING" value="-encoding {{project.java.encoding}}" />
    <option name="MAXIMUM_HEAP_SIZE" value="{{project.java.maximum_heap_size}}" />
  </component>

  <component name="ProjectModuleManager">
    <modules>
      {{#project.modules}}
      <module fileurl="file://{{path}}"
              filepath="{{path}}"/>
      {{/project.modules}}
    </modules>
  </component>

  <component name="ProjectRootManager"
             version="2"
             languageLevel="{{project.java.language_level}}"
             project-jdk-name="{{project.java.jdk}}"
             project-jdk-type="JavaSDK">
    <output url="file://{{project.outdir}}"/>
  </component>

  <component name="ProjectRunConfigurationManager">
    <configuration default="false" name="Debug on port {{project.debug_port}}"
                   type="Remote"
                   factoryName="Remote">
      <option name="USE_SOCKET_TRANSPORT" value="true" />
      <option name="SERVER_MODE" value="false" />
      <option name="SHMEM_ADDRESS" value="javadebug" />
      <option name="HOST" value="localhost" />
      <option name="PORT" value="{{project.debug_port}}" />
      <method />
    </configuration>
  </component>

  {{#project.scala?}}
  <component name="ScalacSettings">
    <option name="USE_FSC"
            {{#project.scala.fsc}}
            value="true"
            {{/project.scala.fsc}}
            {{^project.scala.fsc}}
            value="false"
            {{/project.scala.fsc}}
    />
  </component>
  <component name="libraryTable">
    <library name="scala">
      <CLASSES>
        {{#project.scala.compiler_classpath}}
        <root url="jar://{{.}}!/"/>
        {{/project.scala.compiler_classpath}}
      </CLASSES>
      <JAVADOC />
      <SOURCES />
    </library>
  </component>
  {{/project.scala?}}

  <component name="VcsDirectoryMappings">
    <mapping directory="{{project.root_dir}}" vcs="Git"/>
  </component>

  <component name="CheckStyle-IDEA">
    <option name="configuration">
      <map>
        <entry key="active-configuration" value="file://{{project.root_dir}}/build-support/checkstyle/coding_style.xml:Twitter CheckStyle rules." />
        <entry key="check-test-classes" value="test" />
        <entry key="location-0" value="CLASSPATH:/sun_checks.xml:The default CheckStyle rules." />
        <entry key="location-1" value="file://{{project.root_dir}}/build-support/checkstyle/coding_style.xml:Twitter CheckStyle rules." />
        <entry key="property-1.checkstyle.suppression.files" value="{{project.checkstyle_suppression_files}}"/>
        <entry key="thirdparty-classpath" value="{{project.checkstyle_classpath}}" />
      </map>
    </option>
  </component>

{{#project.extra_components}}
  {{{.}}}

{{/project.extra_components}}
</project>
