<div class="codebase-browse dir-listing monospace">
<div class="breadcrumbs">
<span class="crumb">{{root_parent}}/</span>{{#breadcrumbs}}<span class="crumb"><a href="/browse/{{link_path}}">{{name}}</a>/</span>{{/breadcrumbs}}
</div>
<div class="dir-entries">
{{#entries}}
<a href="/browse/{{link_path}}">{{name}}</a><br>
{{/entries}}
</div>
</div>
