{{! Append some reporting content to a workunit. }}
<div id="__{{content_id}}">{{{content}}}</div>
<script>
    $(function(){
        pants.append('#__{{content_id}}', '#{{workunit_id}}-content');
    });
</script>
