<div {{#class_prefix}}class="{{.}}" {{/class_prefix}}id="{{id}}">
  <div class="{{#class_prefix}}{{.}}-header {{/class_prefix}}toggle-header" id="{{id}}-header">
    <div class="{{#class_prefix}}{{.}}-header-icon {{/class_prefix}}toggle-header-icon" onclick="pants.collapsible.toggle('{{id}}')">
      <i id="{{id}}-icon" class="visibility-icon icon-large {{#initially_open}}icon-caret-down{{/initially_open}}{{^initially_open}}icon-caret-right{{/initially_open}} hidden"></i>
    </div>
    <div class="{{#class_prefix}}{{.}}-header-text {{/class_prefix}}toggle-header-text">
      <div class="timeprefix">
        <span class="timestamp">{{workunit.start_time_string}}</span><span class="timedelta">{{workunit.start_delta_string}}</span>
      </div>
      [<span id="{{id}}-header-text">{{title}}</span>]
      <span class="timer" id="{{id}}-timer"></span>
      {{#icon}}<i class="{{.}}"></i>{{/icon}}
      <span class="aborted nodisplay" id="{{id}}-aborted">ctrl-c</span>
      <span class="unaccounted-time nodisplay" id="{{id}}-unaccounted-time"></span>
    </div>
    {{#spinner}}<div id="{{id}}-spinner"><i class="icon-spinner icon-spin icon-large"></i></div>{{/spinner}}
  </div>
  <div class="{{#class_prefix}}{{.}}-content {{/class_prefix}}toggle-content {{^initially_open}}nodisplay{{/initially_open}}" id="{{id}}-content"></div>
</div>
