
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(src/python/twitter/pants/BUILD:pants-packaged)

from setuptools import setup

setup(**
{   'classifiers': [   'Intended Audience :: Developers',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python'],
    'description': "Twitter's pants build tool.",
    'entry_points': {   'console_scripts': [   'pants = twitter.pants.bin.pants_exe:main']},
    'install_requires': [   'twitter.common.python==0.5.1',
                            'twitter.common.quantity==0.1.3',
                            'twitter.common.confluence==0.1.3',
                            'twitter.common.rwbuf==0.1.0',
                            'pytest-cov==1.6',
                            'twitter.common.config==0.1.3',
                            'setuptools==2.2',
                            'twitter.common.decorators==0.1.3',
                            'python-daemon==1.5.5',
                            'twitter.common.util==0.1.3',
                            'twitter.common.process==0.1.3',
                            'Markdown==2.1.1',
                            'psutil==1.1.2',
                            'pytest==2.5.2',
                            'requests==2.0.0',
                            'elementtree==1.2.7-20070827-preview',
                            'pystache==0.5.3',
                            'Pygments==1.4',
                            'ansicolors==1.0.2',
                            'coverage==3.7.1',
                            'twitter.common.threading==0.1.0'],
    'license': 'Apache License, Version 2.0',
    'name': 'twitter.pants',
    'namespace_packages': ['twitter', 'twitter.pants'],
    'package_data': {   'twitter.pants': [   'docs/pagerank.py',
                                             'docs/conf.py',
                                             'docs/gen.py'],
                        'twitter.pants.java.distribution': [   'SystemProperties.class'],
                        'twitter.pants.reporting': [   'assets/js/prettify_extra_langs/lang-wiki.js',
                                                       'assets/js/prettify.js',
                                                       'assets/js/prettify_extra_langs/lang-ml.js',
                                                       'assets/js/prettify_extra_langs/lang-n.js',
                                                       'templates/tool_invocation_start.mustache',
                                                       'assets/js/jquery-1.8.3.min.js',
                                                       'templates/run.mustache',
                                                       'assets/js/prettify_extra_langs/lang-scala.js',
                                                       'templates/file_content.mustache',
                                                       'assets/js/prettify_extra_langs/lang-clj.js',
                                                       'assets/js/prettify_extra_langs/lang-tex.js',
                                                       'templates/message.mustache',
                                                       'assets/js/prettify_extra_langs/lang-vb.js',
                                                       'assets/font/fontawesome-webfont.eot',
                                                       'assets/font/fontawesome-webfont.svg',
                                                       'assets/css/font-awesome.css',
                                                       'templates/file.mustache',
                                                       'assets/css/pants.css',
                                                       'assets/js/prettify_extra_langs/lang-css.js',
                                                       'templates/run_list.mustache',
                                                       'templates/tool_invocation_end.mustache',
                                                       'assets/js/pants.js',
                                                       'templates/aggregated_timings.mustache',
                                                       'templates/workunit_start.mustache',
                                                       'assets/font/fontawesome-webfont.woff',
                                                       'assets/css/prettify.css',
                                                       'assets/js/prettify_extra_langs/lang-go.js',
                                                       'assets/js/prettify_extra_langs/lang-apollo.js',
                                                       'assets/js/prettify_extra_langs/lang-lua.js',
                                                       'assets/font/FontAwesome.otf',
                                                       'templates/dir.mustache',
                                                       'assets/js/prettify_extra_langs/lang-vhdl.js',
                                                       'assets/js/prettify_extra_langs/lang-xq.js',
                                                       'templates/workunit_end.mustache',
                                                       'assets/js/prettify_extra_langs/lang-hs.js',
                                                       'assets/js/prettify_extra_langs/lang-proto.js',
                                                       'templates/collapsible.mustache',
                                                       'assets/js/prettify_extra_langs/lang-yaml.js',
                                                       'templates/base.mustache',
                                                       'templates/append_to_workunit.mustache',
                                                       'assets/js/prettify_extra_langs/lang-lisp.js',
                                                       'assets/js/prettify_extra_langs/lang-sql.js',
                                                       'assets/font/fontawesome-webfont.ttf'],
                        'twitter.pants.tasks': [   'templates/eclipse/factorypath-3.7.mustache',
                                                   'templates/eclipse/pydevproject-3.7.mustache',
                                                   'templates/eclipse/project-3.7.mustache',
                                                   'templates/jar_publish/ivysettings.mustache',
                                                   'templates/jar_publish/pom.mustache',
                                                   'templates/idea/project-12.mustache',
                                                   'templates/eclipse/org.eclipse.jdt.core.prefs-3.7.mustache',
                                                   'templates/eclipse/debug-launcher-3.7.mustache',
                                                   'templates/idea/module-12.mustache',
                                                   'templates/eclipse/classpath-3.7.mustache',
                                                   'templates/ivy_resolve/ivy.mustache']},
    'package_dir': {   '': 'src'},
    'packages': [   'twitter',
                    'twitter.pants',
                    'twitter.pants.authentication',
                    'twitter.pants.base',
                    'twitter.pants.bin',
                    'twitter.pants.cache',
                    'twitter.pants.commands',
                    'twitter.pants.engine',
                    'twitter.pants.fs',
                    'twitter.pants.goal',
                    'twitter.pants.ivy',
                    'twitter.pants.java',
                    'twitter.pants.java.distribution',
                    'twitter.pants.java.jar',
                    'twitter.pants.net',
                    'twitter.pants.net.http',
                    'twitter.pants.pants_doc',
                    'twitter.pants.process',
                    'twitter.pants.python',
                    'twitter.pants.reporting',
                    'twitter.pants.scm',
                    'twitter.pants.targets',
                    'twitter.pants.tasks',
                    'twitter.pants.tasks.javadoc',
                    'twitter.pants.tasks.javadoc.assets',
                    'twitter.pants.tasks.jvm_compile',
                    'twitter.pants.tasks.jvm_compile.java',
                    'twitter.pants.tasks.jvm_compile.scala',
                    'twitter.pants.tasks.python',
                    'twitter.pants.version'],
    'url': 'https://github.com/twitter/commons',
    'version': '0.0.18',
    'zip_safe': True}
)
