<?xml version="1.0"?>

<!--
=================================================================================================
Copyright 2011 Twitter, Inc.
_________________________________________________________________________________________________
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this work except in compliance with the License.
You may obtain a copy of the License in the LICENSE file, or at:

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
=================================================================================================
-->

<!-- generated by pants! -->
<module type="JAVA_MODULE" version="4">
  <component name="FacetManager">
    {{#module.python?}}
    <facet type="Python" name="Python"/>
    {{/module.python?}}

    {{#module.scala?}}
    <facet type="scala" name="Scala">
      <configuration>
        <option name="compilerLibraryLevel" value="Project"/>
        <option name="compilerLibraryName" value="scala"/>
        <option name="languageLevel" value="{{module.scala.language_level}}" />
        <option name="maximumHeapSize" value="{{module.scala.maximum_heap_size}}"/>
      </configuration>
    </facet>
    {{/module.scala?}}

    {{#module.bash?}}
    <facet type="bash" name="BashSupport"/>
    {{/module.bash?}}
  </component>
  <component name="NewModuleRootManager" inherit-compiler-output="true">
    <exclude-output/>

    <content url="file://{{module.root_dir}}">
    {{#module.content_roots}}
      {{#sources}}
        {{#package_prefix}}
      <sourceFolder url="file://{{module.root_dir}}/{{path}}"
                    isTestSource="{{is_test}}"
                    packagePrefix="{{.}}"/>
        {{/package_prefix}}
        {{^package_prefix}}
      <sourceFolder url="file://{{module.root_dir}}/{{path}}"
                    isTestSource="{{is_test}}"/>
        {{/package_prefix}}
      {{/sources}}
    {{/module.content_roots}}
    {{#module.content_roots}}
      {{#exclude_paths}}
      <excludeFolder url="file://{{module.root_dir}}/{{.}}"/>
      {{/exclude_paths}}
    {{/module.content_roots}}
      <!-- TODO(John Sirois): Kill this exclude hack - its valid to map generated sources under
           .pants.d as generated sources for example - which IDEA 13+ understands. -->
      <excludeFolder url="file://${module.root_dir}/.pants.d" />

      <excludeFolder url="file://$MODULE_DIR$/external-libs" />
      <excludeFolder url="file://$MODULE_DIR$/external-libsources" />
      <excludeFolder url="file://$MODULE_DIR$/internal-libs" />
      <excludeFolder url="file://$MODULE_DIR$/internal-libsources" />
    </content>

    <orderEntry type="inheritedJdk"/>
    <orderEntry type="sourceFolder" forTests="false"/>
    {{#module.scala?}}
    <orderEntry type="library" name="scala" level="project"/>
    {{/module.scala?}}
    <orderEntry type="module-library">
      <library name="internal">
        <CLASSES>
          {{#module.internal_jars}}
          <root url="jar://{{.}}!/" />
          {{/module.internal_jars}}
        </CLASSES>
        <JAVADOC />
        <SOURCES>
          {{#module.internal_source_jars}}
          <root url="jar://{{.}}!/" />
          {{/module.internal_source_jars}}
        </SOURCES>
      </library>
    </orderEntry>
    <orderEntry type="module-library">
      <library name="external">
        <CLASSES>
          {{#module.external_jars}}
          <root url="jar://{{.}}!/" />
          {{/module.external_jars}}
        </CLASSES>
        <JAVADOC>
          {{#module.external_javadoc_jars}}
          <root url="jar://{{.}}!/" />
          {{/module.external_javadoc_jars}}
        </JAVADOC>
        <SOURCES>
          {{#module.external_source_jars}}
          <root url="jar://{{.}}!/" />
          {{/module.external_source_jars}}
        </SOURCES>
      </library>
    </orderEntry>
  </component>

{{#module.extra_components}}
  {{{.}}}

{{/module.extra_components}}
</module>
